#include "PolyDataToImageData.h"

#include <vtkCommand.h>
#include <vtkInformation.h>
#include <vtkInformationVector.h>
#include <vtkObjectFactory.h>
#include <vtkStreamingDemandDrivenPipeline.h>

#include <vtkPolyData.h>
#include <vtkImageData.h>
#include <vtkPolyDataToImageStencil.h>
#include <vtkPointData.h>
#include <vtkImageStencil.h>

vtkStandardNewMacro(PolyDataToImageData);

//----------------------------------------------------------------------------
PolyDataToImageData::PolyDataToImageData()
{
	this->SetNumberOfInputPorts(1);
	this->SetNumberOfOutputPorts(1);
}

//----------------------------------------------------------------------------
PolyDataToImageData::~PolyDataToImageData()
{
}

//----------------------------------------------------------------------------
void PolyDataToImageData::PrintSelf(ostream& os, vtkIndent indent)
{
	this->Superclass::PrintSelf(os, indent);
}

//----------------------------------------------------------------------------
int PolyDataToImageData::ProcessRequest(vtkInformation* request,
	vtkInformationVector** inputVector,
	vtkInformationVector* outputVector)
{
	// If we have no input then we will not generate the output because
	// the user already called StartAppend/Append/EndAppend.
	if (request->Has(vtkDemandDrivenPipeline::REQUEST_DATA_NOT_GENERATED()))
	{
		if (inputVector[0]->GetNumberOfInformationObjects() == 0)
		{
			vtkInformation* outInfo = outputVector->GetInformationObject(0);
			outInfo->Set(vtkDemandDrivenPipeline::DATA_NOT_GENERATED(), 1);
		}
		return 1;
	}
	else if (request->Has(vtkDemandDrivenPipeline::REQUEST_DATA()))
	{
		if (inputVector[0]->GetNumberOfInformationObjects() == 0)
		{
			return 1;
		}
	}

	return this->Superclass::ProcessRequest(request, inputVector, outputVector);
}

//----------------------------------------------------------------------------
int PolyDataToImageData::FillInputPortInformation(
	int vtkNotUsed(port), vtkInformation* info)
{
	info->Set(vtkAlgorithm::INPUT_REQUIRED_DATA_TYPE(), "vtkPolyData");
	return 1;
}

//----------------------------------------------------------------------------
// This is the superclasses style of Execute method.  Convert it into
// an imaging style Execute method.
int PolyDataToImageData::RequestData(
	vtkInformation* vtkNotUsed(request),
	vtkInformationVector **inputVector,
	vtkInformationVector* outputVector)
{
	//Adjust this section depending on number of i/o ports
	//Later on RequestData (RD) happens.
	//During RD each filter examines any inputs it has, then fills in that empty data object with real data.

	vtkInformation* outInfo0 = outputVector->GetInformationObject(0);
	vtkImageData* output0 = vtkImageData::SafeDownCast(
		outInfo0->Get(vtkDataObject::DATA_OBJECT()));

	vtkInformation *inInfo0 = inputVector[0]->GetInformationObject(0);
	vtkPolyData *input0 = vtkPolyData::SafeDownCast(
		inInfo0->Get(vtkDataObject::DATA_OBJECT()));

	if (input0 == nullptr)
	{
		vtkErrorMacro("Input is not defined.");
		return 0;
	}
	
	if (!this->CalculateImageProperties(input0))
		return 0;

	// Set output structure
	vtkInformation* outInfo = outputVector->GetInformationObject(0);

	outInfo->Set(vtkStreamingDemandDrivenPipeline::WHOLE_EXTENT(),
		0, this->Resolution[0],
		0, this->Resolution[1],
		0, this->Resolution[2]);

	outInfo->Set(vtkDataObject::ORIGIN(), this->Origin, 3);
	outInfo->Set(vtkDataObject::SPACING(), this->Spacing, 3);

	vtkDataObject::SetPointDataActiveScalarInfo(outInfo, VTK_FLOAT, 1);
	
	//Desired volume Spacing	
	auto whiteImage = vtkSmartPointer<vtkImageData>::New();    
	whiteImage->SetOrigin(this->Origin);
	whiteImage->SetSpacing(this->Spacing);
	whiteImage->SetExtent(0, this->Resolution[0],
							0, this->Resolution[1],
							0, this->Resolution[2]);

	whiteImage->AllocateScalars(VTK_FLOAT,1);

	//Fill the image with foreground voxels:
	unsigned char inval = 255;
	unsigned char outval = 0;
	vtkIdType count = whiteImage->GetNumberOfPoints();
	for (vtkIdType i = 0; i < count; ++i)
	{
		whiteImage->GetPointData()->GetScalars()->SetTuple1(i, inval);
	}
	 
	//Polygonal data --> image stencil:
	auto pol2stenc = vtkSmartPointer<vtkPolyDataToImageStencil>::New();
	pol2stenc->SetInputData(input0);
	pol2stenc->SetOutputOrigin(this->Origin);
	pol2stenc->SetOutputSpacing(this->Spacing);
	pol2stenc->SetOutputWholeExtent(whiteImage->GetExtent());
	pol2stenc->Update();
	 
	//Cut the corresponding white image and set the background:
	auto imgstenc  = vtkSmartPointer<vtkImageStencil>::New();
	imgstenc->SetInputData(whiteImage);
	imgstenc->SetStencilConnection(pol2stenc->GetOutputPort());
	imgstenc->ReverseStencilOff();
	imgstenc->SetBackgroundValue(outval);
	imgstenc->Update();
	
	//Outputs
	output0->DeepCopy(imgstenc->GetOutput());

	return 1;
}
//----------------------------------------------------------------------------
int PolyDataToImageData::RequestInformation(
	vtkInformation* vtkNotUsed(request),
	vtkInformationVector** inputVector,
	vtkInformationVector* outputVector)
{
	vtkInformation *inInfo0 = inputVector[0]->GetInformationObject(0);
	vtkPolyData* input0 = vtkPolyData::SafeDownCast(
		inInfo0->Get(vtkDataObject::DATA_OBJECT()));

	if (input0 == nullptr)
		return 0;

	if (!this->CalculateImageProperties(input0))
		return 0;

	// Set output structure
	vtkInformation* outInfo = outputVector->GetInformationObject(0);
	
	outInfo->Set(vtkStreamingDemandDrivenPipeline::WHOLE_EXTENT(),
													0, this->Resolution[0],
													0, this->Resolution[1],
													0, this->Resolution[2]);

	outInfo->Set(vtkDataObject::ORIGIN(), this->Origin, 3);
	outInfo->Set(vtkDataObject::SPACING(), this->Spacing, 3);

	vtkDataObject::SetPointDataActiveScalarInfo(outInfo, VTK_FLOAT, 1);

	return 1;
}

int PolyDataToImageData::CalculateImageProperties(vtkPolyData* input)
{
	double bounds[6] = { 0.0 };
	input->GetBounds(bounds);

	// Calculations
	for (unsigned int i = 0; i < 3; i++)
	{
		this->Resolution[i] = std::floor((bounds[i * 2 + 1] - bounds[i * 2]) / (this->VoxelSize[i]));

		if (this->Resolution[i] == 0.0)
		{
			vtkErrorMacro("Resolution component cannot be zero.");
			return 0;
		}
	}

	for (unsigned int i = 0; i < 3; i++)
	{
		this->Origin[i] = bounds[2 * i];
		this->Scale[i] = bounds[i * 2 + 1] - bounds[i * 2];
		this->Spacing[i] = (double)this->Scale[i] / (double)this->Resolution[i];
	}

	return 1;
}

//----------------------------------------------------------------------------
int PolyDataToImageData::RequestUpdateExtent(vtkInformation *,
	vtkInformationVector **inputVector,
	vtkInformationVector *)
{
	// This filter always asks for whole extent downstream. To resample
	// a subset of a structured input, you need to use ExtractVOI.
	vtkInformation *inInfo = inputVector[0]->GetInformationObject(0);
	inInfo->Remove(vtkStreamingDemandDrivenPipeline::UPDATE_EXTENT());
	if (inInfo->Has(vtkStreamingDemandDrivenPipeline::WHOLE_EXTENT()))
	{
		inInfo->Set(vtkStreamingDemandDrivenPipeline::UPDATE_EXTENT(),
			inInfo->Get(vtkStreamingDemandDrivenPipeline::WHOLE_EXTENT()),
			6);
	}

	return 1;
}